# --------------------------------------------------------------------------------------------------------
# Provides example functions for 
#         T. Liebig, Privacy Preserving Centralized Counting of Moving Objects, in AGILE 2015, 
#         F. Bacao, M. Y. Santos, and M. Painho, Eds., Springer International Publishing, 2015, 
#         pp. 91-103. 
#
# Version: GPLv3 http://www.gnu.org/licenses/gpl-3.0.html
#
# June 2015		  								Thomas Liebig
# --------------------------------------------------------------------------------------------------------

cat(" to see a running example just call \n     example() \n")


require("combinat") # binomial coefficient
require("pracma")   # provides gcd()
require("numbers")  # provides isPrime()

example <- function() {
	rabin_seed=42
        l=3
        k=3
	Delta=factorial(l)
        keys=init(7,11,2,l,k)
        publickey=keys$publickey
        shares=keygen(7,11,2,k,l,Delta)
	n = 35
	storedhash=rabin(rabin_seed,n,k=1000,publickey$N^2)
        cat(paste("We assume a single cell, with ",k," moving objects.\n In every round they transmit their count to a server \nwhich aggregates these values.\n\n",sep=""))

        cat(paste("Initializing \n",sep=""))
        cat(paste(" .. generating keys \n",sep=""))
        cat(paste(" .. generating shares \n",sep=""))
        cat(paste(" .. generating hashchain \n",sep=""))
        cat(paste(" DONE \n",sep=""))
        cat(paste("Starting Loop \n",sep=""))

	for (j in 1:10) {
  	   m1=sample(1:100,k)
	   hash=rabin(rabin_seed,n,k=1000-j,publickey$N^2)
	   m=mod(m1*hash,publickey$N^2)
	   cprev=E_shared(m,publickey,Delta,s=2,l)
	   ci=share_combine(cprev,publickey,shares,Delta,s=2,l)
	   m_dec1=D_shared(ci,Delta,publickey,s=2)
	   hash_inv=mod(extGCD(hash,publickey$N^2)[2]+publickey$N^2,publickey$N^2)
	   m_dec=mod(m_dec1*hash_inv,publickey$N^2)
	   cat(paste("\nTimestamp : ",j,"\n",sep=""))
	   for (o in 1:k) {
 	      cat(paste("   Client ",o," computes: \n",sep=""))
 	      cat(paste("      message ",m1[o]," \n",sep=""))
 	      cat(paste("      hash*message ",m[o]," \n",sep=""))
 	      cat(paste("      E(hash*message) ",cprev[o]," \n",sep=""))
	   }
	   cat(paste("   Transmission of encrypted messages to server\n",sep=""))
	   cat(paste("   Server computes:\n",sep=""))
	   cat(paste("      D(share combine):",m_dec1,"\n",sep=""))
	   cat(paste("      D(share combine)*hash^(-1):",m_dec,"\n",sep=""))
	   cat(paste("   homomorphytest: ",paste(m1,collapse=" + "),"=",sum(m1),"=",m_dec," is ",(sum(m1)==m_dec),"\n",sep=""))
	}
        cat(paste(" DONE \n",sep=""))
}

init <- function(p,q,s,l,k) {
   n<-p*q
   nprime<-((p-1)/2)*((q-1)/2)
   g<-n+1
   lambda<-(p-1)*(q-1)/gcd(p-1,q-1)
   d<-chinese(c(0,1),c(lambda,n^s)) 
   s<-2
   publickey<-list()
   privatekey<-list()
   Delta=min(factorial(l),120) 
   publickey$N=n
   publickey$g=g

   privatekey$N=n
   privatekey$d=d
   privatekey$j=1 
   key=list()
   key$publickey=publickey
   key$privatekey=privatekey
   key
}

E <- function(m,publickey) {
  r=23
  t1=modpower(publickey$g,m,publickey$N^(s+1))
  t2=modpower(r,publickey$N^s,publickey$N^(s+1))
  mod(t1*t2,publickey$N^(s+1))
}

D <- function(m,privatekey) {
  i=0

  alpha= modpower(m,privatekey$d,privatekey$N^(s+1))
  for (j in 1:s) {
	t1= (mod(alpha,privatekey$N^(1+j))-1)/n # L(alpha,n,j)
	t2= i
	for (k in 2:j) {
		i=i-1
		t2=mod(t2*i,privatekey$N^j)
		t3=(t2*privatekey$N^(k-1))/factorial(k)
		t1=mod(t1-t3,privatekey$N^j)
	}
	i=t1
  }  
  mod(i*mod(extGCD(privatekey$d,privatekey$N^s)[2]+privatekey$N^s,privatekey$N^s),privatekey$N^s)
}

# example:
# D(E(c(3),publickey),privatekey)

# functions for threshold variant

E_shared <- function(m,publickey,Delta,s,l) {
	r=c()
	for (i in 1:length(m)) {
	# rand=modpower((1+publickey$N),n^s,n^(s+1))

	 weiter=1
	 while (weiter==1) {
	   rand=sample(1:publickey$N^(s+1)-1,1)
	   weiter=0
	#   if (mod(rand,publickey$N)==0) {weiter=1}
	   if (gcd(rand,publickey$N^(s+1))!=1) {weiter=1}
	   if (gcd(rand,publickey$N^s)!=1) {weiter=1}
	   if (gcd(rand,publickey$N)!=1) {weiter=1}
	   if (i>1) {
	     for (j in 1:(i-1)) {
	      if (gcd(r[j],rand)!=1) {weiter=1}
	      }
	    }
	   }
	   r[i]=rand
	   r[i]=1
	}
	 
#	 r=c(302811,339931,38894)
#	 r=c(38894, 339931, 302811) # fixed for test purpose

	c_prev=c()
	for (i in 1:length(m)) {
	  t1=modpower(publickey$g,m[i],publickey$N^(s+1))
	  t2=modpower(r[i],publickey$N^s,publickey$N^(s+1))
	  c_prev[i]=mod(t1*t2,publickey$N^(s+1))
	}
  c_prev
}

share_combine <- function(m,publickey,shares,Delta,s,l) {
	t=1
	for (i in 1:length(m)) {
	  t=mod((t*m[i]),publickey$N^(s+1)) # multiply c_prev
	}
	
	ci=c()
	for (i in 1:l) {
	  ci[i]=modpower(t,(2*Delta*shares[i]),publickey$N^(s+1)) # encrypt t
	}
	 ci
}

D_shared <- function(m,Delta,publickey,s) {
  ci=c()
  cprime=1
  twolambdaS0=c()
  for (i in 0:(length(m)-1)) {
     twolambdaS0[i+1]=Delta
     for (h in 0:(length(m)-1)) {
       if (i!=h) {
          temp1=(-(h+1)) #i
	  td=(i-h)
          temp1=temp1/td
        twolambdaS0[i+1]=twolambdaS0[i+1]*temp1
       }
     }
  }
  cprime=1
  for (i in 0:(length(m)-1)) {
     twolambdaS0[i+1]=mod(twolambdaS0[i+1]+publickey$N^(s+1),publickey$N^(s+1))
     cprime=cprime*modpower(m[i+1],(2*twolambdaS0[i+1]),publickey$N^(s+1))
     cprime=mod(cprime,publickey$N^(s+1))
  }

  {
  i_jPrev=cprime-1
  i_jPrev=i_jPrev/publickey$N
  i_jPrev=mod(i_jPrev,(publickey$N)^2)
  if (s==1) {i_jCur=i_jPrev} 
  for (j in (2:s)) {
    i_jCur=mod(cprime,publickey$N^(j+1))-1
    i_jCur=i_jCur/publickey$N
    for (k in (2:s)) {
      temp=nCm(i_jPrev,k)
      temp=temp %% (publickey$N^j)
      temp=temp*publickey$N^(k-1)
      temp=temp %% (publickey$N^j)
      i_jCur=i_jCur-temp
    }
    i_jPrev=i_jCur
  }
  temp=mod(extGCD(4*Delta^2,publickey$N^s)[2]+publickey$N^s,publickey$N^s)
  i_jCur=mod(round(i_jCur)+(publickey$N^s),publickey$N^s)
  res=mod((i_jCur*temp) + (publickey$N^s),publickey$N^s)
  }

  # in case d of the privatekey would be known it is equivalent to:
  # res=mod(D(round(cprime),privatekey)*mod(round(extGCD(4*Delta^2,n^s)[2]+n^s),round(n^s)),n^s)  
  res
}


keygen <-function(p,q,s,k,l,Delta) {
# shares the private d on several parties
	n<-p*q
	nprime<-((p-1)/2)*((q-1)/2)
	g<-n+1
	lambda<-(p-1)*(q-1)/gcd(p-1,q-1)
	d<-chinese(c(0,1),c(lambda,n^s)) 
	f=0
	m=((p-1)/2)*((q-1)/2)
	v=10
	v_share=c()

	a=c()
	for(i in 1:(k)) {
	 if (i==1) {a[i]=d} else 
	 { 
	   mnpows=m*n^s
	   a[i]=sample(1:mnpows,1) 
	 }
	}

	share=c()
	for (i in 1:l) {
	  share[i]=0
	  for (j in 1:(k)) {
	    temp=((i)^(j-1))*a[j]
	    share[i]=share[i]+temp
	    share[i]=mod(share[i],(n^s)*nprime)
	  }
	}
	share
}

rabin <- function (m,n,k=1,crypt_n) {
# Rabin function is supposed to be a one-way function
#      n is a product of primes p*q
#      k is exponent
	res=m
        for (i in 1:k) {
	  res=(mod(res*res,n))
        }
	while (gcd(res,crypt_n)!=1) { # ensure that there xists a multiplicative inverse to rabin mod crypt_n
	  res=(res+1)
	}
	return(res)
}
